from .kicad import parse_sexpr, build_sexpr

from .lists import ListAt, ListId, ListUuid, ListLibId, ListUnit, ListInBom, ListOnBoard
from .lists import ListSize, ListFont, ListEffects, ListProperty, ListSymbolPin, ListSymbol

from .schematic import Schematic



from .sexpr import SExpr
from .lists import ListSymbol


class Schematic(SExpr):
    def __init__(self, data):
        super().__init__(data, required_tag='kicad_sch', required_atoms=('string',),
                         required_lists=('version', 'generator',
                                         'uuid', 'paper',),
                         optional_lists=('lib_symbols', 'junction[]', 'label[]', 'text[]', 'wire[]', 'no_connect[]', 'global_label[]',
                                         'sheet[]', 'sheet_instances', 'symbol_instances', 'symbol[]',),
                         )
        self._validate_or_raise("invalid initial data")

    @property
    def symbols(self) -> ():
        symbols = [ListSymbol(x) for x in self.all('symbol')]
        return tuple(symbols)



from john import TestCase

from .base import SExprBase
from .kicad import parse_sexpr

symbol_text = """
  (symbol (lib_id "power:GND") (at 168.91 93.98 0) (unit 1)
    (in_bom yes) (on_board yes)
    (uuid 18af9582-7c92-46f1-9603-90a0db4828ae)
    (property "Reference" "#PWR08" (id 0) (at 168.91 100.33 0)
      (effects (font (size 1.27 1.27)) hide)
    )
    (property "Value" "GND" (id 1) (at 169.037 98.3742 0))
    (property "Footprint" "" (id 2) (at 168.91 93.98 0)
      (effects (font (size 1.27 1.27)) hide)
    )
    (property "Datasheet" "" (id 3) (at 168.91 93.98 0)
      (effects (font (size 1.27 1.27)) hide)
    )
    (pin "1" (uuid 0462adc4-54a7-4c18-967e-f6d78a20207f))
  )
"""

property_text = """
    (property "Value" "2.2K 1%" (id 1) (at 179.578 123.19 0)
      (effects (font (size 1.27 1.27)) (justify left))
    )
"""


class KiFabSExprSubClass(SExprBase):
    pass


class TestInit(TestCase):
    def setUp(self):
        SExprBase._allow_instantiate_base = True
        SExprBase.allow_empty = True

    def test_instantiate_raises_if_not_allowed(self):
        SExprBase._allow_instantiate_base = False

        with self.assertRaisesAny():
            SExprBase(['ZZ9PZA'])

    def test_instantiate_does_not_raise_if_allowed(self):
        SExprBase._allow_instantiate_base = True

        with self.assertDoesNotRaise():
            SExprBase(['ZZ9PZA'])

    def test_instantiate_does_not_raise_if_not_allowed(self):
        SExprBase._allow_instantiate_base = False

        with self.assertDoesNotRaise():
            tag = 'ZZ9PZA'
            data = [tag]
            sut = KiFabSExprSubClass(data)

            self.assertIsInstance(sut, KiFabSExprSubClass)

            expected = tag

            actual = sut.tag

            self.assertEqual(actual, expected)

    def test_throws_with_None_and_not_allow_empty(self):
        SExprBase.allow_empty = False

        with self.assertRaisesAny():
            SExprBase(None)

    def test_does_not_throw_with_None_and_allow_empty(self):
        SExprBase.allow_empty = True

        with self.assertDoesNotRaise():
            SExprBase(None)

    def test_throws_with_invalid_data(self):
        invalid_data = "(ZZ9PZA)"

        with self.assertRaisesAny():
            SExprBase(invalid_data)

    def test_is_empty_when_empty(self):
        sut = SExprBase(None)

        self.assertTrue(sut.is_empty)

    def test_is_empty_when_not_empty(self):
        data = ['ZZ9PZA']

        sut = SExprBase(data)

        self.assertFalse(sut.is_empty)


class TestTag(TestCase):
    def test_tag_with_valid_tag(self):
        expected = 'ZZ9PZA'

        actual = SExprBase([expected]).tag

        self.assertEqual(actual, expected)

    def test_does_not_raise_without_tag(self):
        with self.assertDoesNotRaise():
            SExprBase([]).tag

    def test_invalid_tag_returns_none(self):
        self.assertIsNone(SExprBase([1729]).tag)


class TestAtoms(TestCase):
    def setUp(self):
        SExprBase._allow_instantiate_base = True
        SExprBase.allow_empty = False

        self.symbol = SExprBase(data=parse_sexpr(symbol_text))

    def test_atoms(self):
        expected = ('symbol',)

        actual = self.symbol._atoms

        self.assertTupleEqual(actual, expected)

    def test_add_atom_None_raises(self):
        with self.assertRaisesAny():
            self.symbol._add_atom(None)

    def test_add_atom_list_raises(self):
        with self.assertRaisesAny():
            self.symbol._add_atom([])

    def test_add_atom_tuple_raises(self):
        with self.assertRaisesAny():
            self.symbol._add_atom(())

    def test_add_atom_string_does_not_raise(self):
        with self.assertDoesNotRaise():
            self.symbol._add_atom("ZZ9PZA")

    def test_add_atom_int_does_not_raise(self):
        with self.assertDoesNotRaise():
            self.symbol._add_atom(42)

    def test_add_atom_float_does_not_raise(self):
        with self.assertDoesNotRaise():
            self.symbol._add_atom(3.14)

    def test_add_atom_increases_number_of_atoms(self):
        expected = len(self.symbol._atoms) + 1

        self.symbol._add_atom('ZZ9PZA')

        actual = len(self.symbol._atoms)

        self.assertEqual(actual, expected)

    def test_add_atom(self):
        atom = 'ZZ9PZA'

        self.symbol._add_atom(atom)

        expected = ('symbol', atom)

        actual = self.symbol._atoms

        self.assertTupleEqual(actual, expected)

    def test_atom_indices_one_list(self):
        sut = SExprBase(['ZZ9PZA', [], 'A'])

        expected = (0, 2,)

        actual = sut._atom_indicies

        self.assertSequenceEqual(actual, expected)

    def test_set_atom_out_of_range_raises(self):
        sut = SExprBase(['a', 'b'])

        with self.assertRaisesAny():
            sut._set_atom(2, 'a')

    def test_set_atom_wrong_type_raises(self):
        sut = SExprBase(['a', 'b'])

        with self.assertRaisesAny():
            sut._set_atom(1, [])

    def test_set_atom_invalid_index_raises(self):
        with self.assertRaisesAny():
            SExprBase(['a', 'b'])._set_atom('a', 'b')

    def test_set_atom_sets_value(self):
        data = ['a', [], 'b']

        expected = 'c'

        sut = SExprBase(data)
        sut._set_atom(1, expected)

        actual = sut._atoms[1]

        self.assertEqual(actual, expected)

    def test_remove_atom_decreases_number_of_atoms(self):
        expected = len(self.symbol._atoms) - 1

        self.symbol._remove_atom('symbol')

        actual = len(self.symbol._atoms)

        self.assertEqual(actual, expected)

    def test_remove_missing_atom_does_not_raise(self):
        with self.assertDoesNotRaise():
            self.symbol._remove_atom('ZZ9PZA')

    # todo foss: more than one atom


class TestLists(TestCase):
    def setUp(self):
        SExprBase._allow_instantiate_base = True
        SExprBase.allow_empty = True

        self.symbol = SExprBase(data=parse_sexpr(symbol_text))

    def test_lists(self):
        expected = (
            ['lib_id', 'power:GND'],
            ['at', 168.91, 93.98, 0],
            ['unit', 1],
            ['in_bom', 'yes'],
            ['on_board', 'yes'],
            ['uuid', '18af9582-7c92-46f1-9603-90a0db4828ae'],
            ['property', 'Reference', '#PWR08', ['id', 0], ['at', 168.91, 100.33, 0], [
                'effects', ['font', ['size', 1.27, 1.27]], 'hide']],
            ['property', 'Value', 'GND', ['id', 1], ['at', 169.037, 98.3742, 0]],
            ['property', 'Footprint', '', ['id', 2], ['at', 168.91, 93.98, 0],
                ['effects', ['font', ['size', 1.27, 1.27]], 'hide']],
            ['property', 'Datasheet', '', ['id', 3], ['at', 168.91, 93.98, 0],
                ['effects', ['font', ['size', 1.27, 1.27]], 'hide']],
            ['pin', '1', ['uuid', '0462adc4-54a7-4c18-967e-f6d78a20207f']]
        )

        actual = self.symbol._lists

        self.assertTupleEqual(actual, expected)

    def test_find_lists_returns_tuple(self):
        keys = ("property", "Value")

        actual = self.symbol._find_lists(keys)

        self.assertIsInstance(actual, tuple)
        self.assertEqual(len(actual), 1)

    def test_find_lists_returns_correct_types(self):
        ll = self.symbol._find_lists(("property", "Value"))
        actual = ll[0]

        self.assertEqual(len(actual), 5)
        self.assertIsInstance(actual[0], str)
        self.assertIsInstance(actual[1], str)
        self.assertIsInstance(actual[2], str)
        self.assertIsInstance(actual[3], list)
        self.assertIsInstance(actual[4], list)

    def test_find_lists_returns_correct_data(self):
        expected = ['property', 'Value', 'GND', [
            'id', 1], ['at', 169.037, 98.3742, 0]]

        actual = self.symbol._find_lists(("property", "Value"))[0]

        self.assertListEqual(actual, expected)

    def test_add_None_list_raises(self):
        with self.assertRaisesAny():
            self.symbol._add_list(None)

    def test_add_empty_list_raises(self):
        with self.assertRaisesAny():
            self.symbol._add_list([])

    def test_add_one_element_list_raises(self):
        with self.assertRaisesAny():
            self.symbol._add_list([None])

    def test_add_two_element_list_without_name_raises(self):
        with self.assertRaisesAny():
            self.symbol._add_list([None, None])

    def test_add_two_element_list_with_non_string_name_raises(self):
        with self.assertRaisesAny():
            self.symbol._add_list([2, None])

    def test_add_two_element_list_with_empty_name_raises(self):
        with self.assertRaisesAny():
            self.symbol._add_list(['', None])

    def test_add_two_element_list_with_name_does_not_raise(self):
        with self.assertDoesNotRaise():
            self.symbol._add_list(['ZZ9PZA', None])

    def test_add_list_increases_number_of_lists(self):
        expected = len(self.symbol._lists) + 1

        self.symbol._add_list(['A', 'B'])

        actual = len(self.symbol._lists)

        self.assertEqual(actual, expected)

    def test_add_list_adds_list(self):
        ll = ["ZZ9PZA", 42]

        self.symbol._add_list(ll)

        expected = ll

        actual = self.symbol._lists[11]

        self.assertListEqual(actual, expected)

    def test_remove_nonexistent_list_raises(self):
        with self.assertRaisesAny():
            self.symbol._remove_list(['a', 'b'])

    def test_remove_list_decreases_number_of_lists(self):
        expected = len(self.symbol._lists) - 1

        ll = self.symbol._lists[1]

        self.symbol._remove_list(ll)

        actual = len(self.symbol._lists)

        self.assertEqual(actual, expected)

    def test_remove_list_removes_list(self):
        expected = self.symbol._lists[2]

        self.symbol._remove_list(self.symbol._lists[1])

        actual = self.symbol._lists[1]

        self.assertListEqual(actual, expected)


class SanityChecks(TestCase):
    def setUp(self):
        SExprBase._allow_instantiate_base = False
        SExprBase.allow_empty = False

        self.symbol = KiFabSExprSubClass(data=parse_sexpr(symbol_text))

    def test_update_list_works(self):
        value = self.symbol._find_lists(("property", "Value"))[0]

        expected = "3v3"

        value[2] = expected

        actual = self.symbol._data[8][2]

        self.assertEqual(actual, expected)

    def test_overridden_new_has_not_done_anything_weird(self):
        tag = 'ZZ9PZA'
        data = [tag]
        sut = KiFabSExprSubClass(data)

        self.assertIsInstance(sut, KiFabSExprSubClass)

        expected = tag

        actual = sut.tag

        self.assertEqual(actual, expected)



import math

from .sexpr import SExpr


class ListAt(SExpr):  # (at 168.91 100.33 0)
    def __init__(self, data):
        super().__init__(data, required_tag='at', required_atoms=(
            'string', 'number', 'number',), optional_atoms=('number',))
        self._validate_or_raise("invalid initial data")

    @property
    def x(self) -> str:
        return self._atoms[1]

    @x.setter
    def x(self, v: float):
        self._set_atom(1, v)
        self._validate_or_raise()

    @property
    def y(self) -> str:
        return self[2]

    @y.setter
    def y(self, v: float):
        self._set_atom(2, v)
        self._validate_or_raise()

    @property
    def angle(self) -> str:
        return self[3]

    @angle.setter
    def angle(self, v: float):
        self._set_atom(3, v)
        self._validate_or_raise()

    def distance_from(self, other) -> float:
        x = other.x - self.x
        y = other.y - self.y
        return math.sqrt(x * x + y * y)


class ListId(SExpr):  # (id 0)
    def __init__(self, data):
        super().__init__(data, required_tag='id', required_atoms=('string', 'number',))
        self._validate_or_raise("invalid initial data")

    @property
    def id(self) -> int:
        return self._atoms[1]

    @id.setter
    def id(self, v: int):
        self._set_atom(1, v)
        self._validate_or_raise()


class ListUuid(SExpr):  # (uuid 18af9582-7c92-46f1-9603-90a0db4828ae)
    def __init__(self, data):
        super().__init__(data, required_tag='uuid', required_atoms=('string', 'string',))
        self._validate_or_raise("invalid initial data")

    @property
    def uuid(self) -> int:
        return self._atoms[1]

    @uuid.setter
    def uuid(self, v: int):
        self._set_atom(1, v)
        self._validate_or_raise()


class ListLibId(SExpr):  # (lib_id "power:GND")
    def __init__(self, data):
        super().__init__(data, required_tag='lib_id', required_atoms=('string', 'string',))
        self._validate_or_raise("invalid initial data")

    @property
    def lib_id(self) -> int:
        return self._atoms[1]

    @lib_id.setter
    def lib_id(self, v: int):
        self._set_atom(1, v)
        self._validate_or_raise()


class ListUnit(SExpr):  # (unit 1)
    def __init__(self, data):
        super().__init__(data, required_tag='unit', required_atoms=('string', 'number',))
        self._validate_or_raise("invalid initial data")

    @property
    def unit(self) -> int:
        return self._atoms[1]

    @unit.setter
    def unit(self, v: int):
        self._set_atom(1, v)
        self._validate_or_raise()


class ListInBom(SExpr):  # (in_bom yes)
    def __init__(self, data):
        super().__init__(data, required_tag='in_bom', required_atoms=('string', 'string',))
        self._validate_or_raise("invalid initial data")

    @property
    def in_bom(self) -> str:
        return self._atoms[1]

    @in_bom.setter
    def in_bom(self, v: str):
        self._set_atom(1, v)
        self._validate_or_raise()


class ListOnBoard(SExpr):  # (on_board yes)
    def __init__(self, data):
        super().__init__(data, required_tag='on_board', required_atoms=('string', 'string',))
        self._validate_or_raise("invalid initial data")

    @property
    def on_board(self) -> str:
        return self._atoms[1]

    @on_board.setter
    def on_board(self, v: str):
        self._set_atom(1, v)
        self._validate_or_raise()


class ListSize(SExpr):  # (size 1.27 1.27)
    def __init__(self, data):
        super().__init__(data, required_tag='size',
                         required_atoms=('string', 'number', 'number',))
        self._validate_or_raise("invalid initial data")

    @property
    def x(self) -> float:
        return self._atoms[1]

    @x.setter
    def x(self, v: float):
        self._set_atom(1, v)
        self._validate_or_raise()

    @property
    def y(self) -> float:
        return self._atoms[2]

    @y.setter
    def y(self, v: float):
        self._set_atom(2, v)
        self._validate_or_raise()


class ListFont(SExpr):  # (font (size 1.27 1.27))
    def __init__(self, data):
        super().__init__(data, required_tag='font',
                         required_atoms=('string',), required_lists=('size',))
        self._validate_or_raise("invalid initial data")


class ListEffects(SExpr):  # (effects (font (size 1.27 1.27)) hide)
    def __init__(self, data):
        super().__init__(data, required_tag='effects', required_atoms=(
            'string',), optional_atoms=('string',), required_lists=('font',))
        self._validate_or_raise("invalid initial data")


class ListProperty(SExpr):
    # (property "Reference" "#PWR08" (id 0) (at 168.91 100.33 0)
    #   (effects (font (size 1.27 1.27)) hide)
    # )
    def __init__(self, data):
        super().__init__(data, required_tag='property',
                         required_atoms=('string', 'string', 'string'),
                         optional_lists=('id', 'at', 'effects',))
        self._validate_or_raise("invalid initial data")

    @property
    def key(self) -> str:
        return self._atoms[1]

    @property
    def value(self) -> str:
        return self._atoms[2]

    @value.setter
    def value(self, value):
        self[self._atom_indicies[2]] = value


class ListSymbolPin(SExpr):  # (pin "1" (uuid 0462adc4-54a7-4c18-967e-f6d78a20207f))
    def __init__(self, data):
        super().__init__(data, required_tag='pin', required_atoms=(
            'string', 'string',), required_lists=('uuid',))
        self._validate_or_raise("invalid initial data")

    @property
    def key(self) -> str:
        return self._atoms[1]


class ListSymbol(SExpr):
    #   (symbol (lib_id "power:GND") (at 168.91 93.98 0) (unit 1)
    #     (in_bom yes) (on_board yes)
    #     (uuid 18af9582-7c92-46f1-9603-90a0db4828ae)
    #     (property "Reference" "#PWR08" (id 0) (at 168.91 100.33 0)
    #       (effects (font (size 1.27 1.27)) hide)
    #     )
    #     (property "Value" "GND" (id 1) (at 169.037 98.3742 0))
    #     (property "Footprint" "" (id 2) (at 168.91 93.98 0)
    #       (effects (font (size 1.27 1.27)) hide)
    #     )
    #     (property "Datasheet" "" (id 3) (at 168.91 93.98 0)
    #       (effects (font (size 1.27 1.27)) hide)
    #     )
    #     (pin "1" (uuid 0462adc4-54a7-4c18-967e-f6d78a20207f))
    #   )
    def __init__(self, data):
        super().__init__(data, required_tag='symbol', required_atoms=('string',),
                         required_lists=('lib_id', 'at', 'unit', 'in_bom', 'on_board', 'uuid', 'property[]', 'pin[]'))
        self._validate_or_raise("invalid initial data")

    @property
    def properties(self) -> ():
        props = [ListProperty(x) for x in self._find_lists('property')]
        return tuple(props)

    @property
    def pins(self) -> ():
        pins = [ListSymbolPin(x) for x in self._find_lists('pin')]
        return tuple(pins)

    def _property(self, key: str) -> ListProperty:
        selected = [x for x in self.properties if x.key == key]

        return None if 0 == len(selected) else selected[0]

    @property
    def datasheet(self) -> ListProperty:
        return self._property('Datasheet')

    @property
    def footprint(self) -> ListProperty:
        return self._property('Footprint')

    @property
    def reference(self) -> ListProperty:
        return self._property('Reference')

    @property
    def value(self) -> ListProperty:
        return self._property('Value')

    @property
    def lib_id(self) -> ListLibId:
        return ListLibId(self["lib_id"])

    @property
    def at(self) -> ListAt:
        return ListAt(self['at'])

    @property
    def unit(self) -> ListUnit:
        return ListUnit(self['unit'])

    @property
    def in_bom(self) -> ListInBom:
        return ListInBom(self['in_bom'])

    @property
    def on_board(self) -> ListOnBoard:
        return ListOnBoard(self['on_board'])

    @property
    def uuid(self) -> ListUuid:
        return ListUuid(self['uuid'])

    def add_property(self, prop: ListProperty) -> None:
        self.add_list(prop._data)

    def get_property(self, key: str) -> ListProperty:
        return self._property(key)



from john import TestCase

from .lists import ListAt, ListId, ListUuid, ListLibId, ListUnit, ListInBom, ListOnBoard
from .lists import ListSize, ListFont, ListEffects, ListProperty, ListSymbolPin, ListSymbol

# (property "Reference" "#PWR08" (id 0) (at 168.91 100.33 0)
#       (effects (font (size 1.27 1.27)) hide)
#     )

# (property "Value" "GND" (id 1) (at 169.037 98.3742 0))

#     (property "Footprint" "" (id 2) (at 168.91 93.98 0)
#       (effects (font (size 1.27 1.27)) hide)
#     )
#     (property "Datasheet" "" (id 3) (at 168.91 93.98 0)
#       (effects (font (size 1.27 1.27)) hide)
#     )
#     (pin "1" (uuid 0462adc4-54a7-4c18-967e-f6d78a20207f))


class TestAt(TestCase):
    def test_init_with_wrong_tag(self):
        with self.assertRaisesAny():
            ListAt(['not_at'])

    def test_init_with_missing_required_Y_atom(self):
        with self.assertRaisesAny():
            ListAt(['at'], 1)

    def test_init_with_required_X_Y(self):
        with self.assertDoesNotRaise():
            ListAt(['at', 1, 2])

    def test_init_with_optional_angle(self):
        with self.assertDoesNotRaise():
            ListAt(['at', 1, 2, 3])


class TestAtdistance_from(TestCase):
    def test_valid_parm_does_raise(self):
        sut = ListAt(['at', 0, 0])

        with self.assertDoesNotRaise():
            sut.distance_from(ListAt(['at', 0, 0]))

    def test_distance(self):
        sut = ListAt(['at', 1, 1])

        actual = sut.distance_from(ListAt(['at',  1 + 3, 1 + 4]))

        self.assertEqual(actual, 5)


class TestAtProperties(TestCase):
    def setUp(self):
        self.sut = ListAt(['at', 42, 1729, 3.14])

    def test_tag(self):
        self.assertEqual(self.sut.tag, 'at')

    def test_getters(self):
        self.assertEqual(self.sut.x, 42)
        self.assertEqual(self.sut.y, 1729)
        self.assertEqual(self.sut.angle, 3.14)

    def test_set_x(self):
        self.sut.x = 13
        self.assertEqual(self.sut.tag, 'at')
        self.assertEqual(self.sut.x, 13)
        self.assertEqual(self.sut.y, 1729)
        self.assertEqual(self.sut.angle, 3.14)

    def test_set_y(self):
        self.sut.y = 27
        self.assertEqual(self.sut.tag, 'at')
        self.assertEqual(self.sut.x, 42)
        self.assertEqual(self.sut.y, 27)
        self.assertEqual(self.sut.angle, 3.14)

    def test_set_angle(self):
        self.sut.angle = 0
        self.assertEqual(self.sut.tag, 'at')
        self.assertEqual(self.sut.x, 42)
        self.assertEqual(self.sut.y, 1729)
        self.assertEqual(self.sut.angle, 0)

    def test_set_x_invalid_raises(self):
        with self.assertRaisesAny():
            self.sut.x = 'a'

    def test_set_y_invalid_raises(self):
        with self.assertRaisesAny():
            self.sut.y = 'a'

    def test_set_angle_invalid_raises(self):
        with self.assertRaisesAny():
            self.sut.angle = 'a'


class TestUuid(TestCase):
    def setUp(self):
        self.uuid = '18af9582-7c92-46f1-9603-90a0db4828ae'

    def test_init_with_wrong_tag(self):
        with self.assertRaisesAny():
            ListUuid(['not_uuid'])

    def test_init_with_missing_required_uuid_atom(self):
        with self.assertRaisesAny():
            ListUuid(['uuid'],)

    def test_init_with_required_uuid(self):
        with self.assertDoesNotRaise():
            ListUuid(['uuid', self.uuid])


class TestLibId(TestCase):
    def test_init(self):
        with self.assertDoesNotRaise():
            ListLibId(['lib_id', 'power:GND'])


class TestUnit(TestCase):
    def test_init(self):
        with self.assertDoesNotRaise():
            ListUnit(['unit', 1])


class TestInBom(TestCase):
    def test_init(self):
        with self.assertDoesNotRaise():
            ListInBom(['in_bom', 'yes'])


class TestOnBoard(TestCase):
    def test_init(self):
        with self.assertDoesNotRaise():
            ListOnBoard(['on_board', 'yes'])


class TestSize(TestCase):
    def test_init(self):
        with self.assertDoesNotRaise():
            ListSize(['size', 3.14, 1729])


class TestFont(TestCase):
    def test_init(self):
        with self.assertDoesNotRaise():
            ListFont(['font', ['size', 3.14, 1729]])


class TestEffects(TestCase):
    def test_init(self):
        with self.assertDoesNotRaise():
            ListEffects(['effects', ['font', ['size', 3.14, 1729]], 'hide'])


class TestProperty(TestCase):
    def test_init(self):
        with self.assertDoesNotRaise():
            ListProperty(['property', 'Reference', '#PWR08',
                          ['id', 1], ['at', 168.91, 100.33, 0],
                          ['effects', ['font', ['size', 3.14, 1729]], 'hide']])

    def test_key(self):
        sut = ListProperty(['property', 'Reference', '#PWR08',
                            ['id', 1], ['at', 168.91, 100.33, 0],
                            ['effects', ['font', ['size', 3.14, 1729]], 'hide']])

        self.assertEqual("Reference", sut.key)


class TestSymbolPin(TestCase):
    def test_init(self):
        with self.assertDoesNotRaise():
            ListSymbolPin(
                ['pin', '1', ['uuid', '0462adc4-54a7-4c18-967e-f6d78a20207f']])

    def test_key(self):
        sut = ListSymbolPin(
            ['pin', '1', ['uuid', '0462adc4-54a7-4c18-967e-f6d78a20207f']])
        self.assertEqual('1', sut.key)


symbol_data = ['symbol',
               ['lib_id', 'power:GND'],
               ['at',  168.91, 93.98, 0],
               ['unit', 1],
               ['in_bom', 'yes'],
               ['on_board', 'yes'],
               ['uuid', '18af9582-7c92-46f1-9603-90a0db4828ae'],
               ['property', 'Reference', '#PWR08',
                ['id', 1], ['at', 168.91, 100.33, 0],
                   ['effects', ['font', ['size', 3.14, 1729]], 'hide']],
               ['property', 'Value', 'GND', ['id', 1], ['at', 169.037, 98.3742, 0]],
               ['property', 'Footprint', '', ['id', 2], ['at', 168.91, 93.98, 0],
                ['effects', ['font', ['size', 3.14, 1729]], 'hide']],
               ['property', 'Datasheet', '', ['id', 3], ['at', 168.91, 93.98, 0],
                ['effects', ['font', ['size', 3.14, 1729]], 'hide']],
               ['pin', '1', ['uuid', '0462adc4-54a7-4c18-967e-f6d78a20207f']],
               ['pin', '2', ['uuid', 'ab6930b8-e44b-4fa6-a4a5-0f73c81f4f14']],
               ]


class TestSymbol(TestCase):
    def setUp(self):
        self.sut = ListSymbol(symbol_data)

    def test_init(self):
        with self.assertDoesNotRaise():
            ListSymbol(self.sut)


class TestSymbolProperties(TestCase):
    def setUp(self):
        self.sut = ListSymbol(symbol_data)

    def test_properies_is_properties_list(self):
        self.assertIsInstance(self.sut.properties, tuple)

    def test_properties_has_correct_length(self):
        self.assertEqual(4, len(self.sut.properties))

    def test_properties_contains_properties(self):
        for prop in self.sut.properties:
            self.assertIsInstance(prop, ListProperty)

    def test_properties_has_correct_properties(self):
        expected = ('Reference', 'Value', 'Footprint', 'Datasheet',)

        actual = tuple([x.key for x in self.sut.properties])

        self.assertSequenceEqual(actual, expected)

    def test_get_property_by_key(self):
        expected = "Datasheet"

        actual = self.sut.get_property(expected)

        self.assertEqual(actual.key, expected)

    def test_missing_property_returns_None(self):
        self.assertIsNone(self.sut._property("Missing"))

    def test_value(self):
        self.assertEqual('#PWR08', self.sut._property("Reference").value)

    def test_write_value(self):
        expected = 'NEW_FOOTPRINT'

        self.sut._property("Footprint").value = expected

        actual = self.sut._data[9][2]

        self.assertEqual(actual, expected)


class TestSymbolLists(TestCase):
    def setUp(self):
        self.sut = ListSymbol(symbol_data)

    def test_lib_id(self):
        self.assertEqual(self.sut.lib_id.lib_id, "power:GND")

    def test_at(self):
        self.assertEqual(self.sut.at.x, 168.91)
        self.assertEqual(self.sut.at.y, 93.98)
        self.assertEqual(self.sut.at.angle, 0)

    def test_unit(self):
        self.assertEqual(1, self.sut.unit.unit)

    def test_in_bom(self):
        self.assertEqual('yes', self.sut.in_bom.in_bom)

    def test_on_board(self):
        self.assertEqual('yes', self.sut.on_board.on_board)

    def test_uuid(self):
        self.assertEqual(
            '18af9582-7c92-46f1-9603-90a0db4828ae', self.sut.uuid.uuid)

    def test_reference(self):
        self.assertEqual("Reference", self.sut.reference.key)

    def test_value(self):
        self.assertEqual("Value", self.sut.value.key)
        self.assertEqual("GND", self.sut.value.value)

    def test_footprint(self):
        self.assertEqual("Footprint", self.sut.footprint.key)

    def test_datasheet(self):
        self.assertEqual("Datasheet", self.sut.datasheet.key)

    def test_pins(self):
        self.assertIsInstance(self.sut.pins, tuple)

    def test_pins_by_key(self):
        self.assertEqual('2', self.sut.pins[1].key)



from john import TestCase

from .schematic import Schematic


class TestSchematic(TestCase):
    def test_init_with_None_raises(self):
        with self.assertRaisesAny():
            Schematic()

    def test_init_with_wrong_tag_raises(self):
        data = ['ZZ9PZA', ['version', '20211123'], ['generator', 'eeschema']]

        with self.assertRaisesAny():
            Schematic(data)



from typing import Any

from .base import SExprBase


class SExprEmpty(SExprBase):
    def __init__(self):
        super().__init__(None)


class SExpr(SExprBase):
    def __init__(self, data, required_tag=None,
                 required_atoms=None, optional_atoms=None,
                 required_lists=None, optional_lists=None):

        self._messages = []

        self._required_tag = required_tag
        self._required_atoms = required_atoms
        self._optional_atoms = optional_atoms
        self._required_lists = required_lists
        self._optional_lists = optional_lists

        if self._required_atoms is not None:
            if not isinstance(self._required_atoms, tuple):
                self._messages.append('required atoms must be tuple')
                raise RuntimeError()

            for required_type in self._required_atoms:
                if 'string' != required_type and 'number' != required_type:
                    self._messages.append(
                        f"invalid atom type '{required_type}'")
                    raise RuntimeError()

        if self._optional_atoms is not None:
            if not isinstance(self._optional_atoms, tuple):
                raise RuntimeError("optional atoms must be tuple")

            for type in self._optional_atoms:
                if 'string' != type and 'number' != type:
                    raise RuntimeError(f"invalid atom type '{type}'")

        if self._required_lists is not None:
            if not isinstance(self._required_lists, tuple):
                raise RuntimeError("required lists must be tuple")

            if 0 != len([x for x in self._required_lists if not isinstance(x, str)]):
                raise RuntimeError("required lists must contain strings")

        if self._optional_lists is not None:
            if not isinstance(self._optional_lists, tuple):
                raise RuntimeError("optional lists must be tuple")

            if 0 != len([x for x in self._optional_lists if not isinstance(x, str)]):
                raise RuntimeError("optional lists must contain strings")

        if data is None:
            raise RuntimeError("data must not be None")

        if not isinstance(data, list) and not isinstance(data, SExpr):
            self._error("data must be a list")
            raise RuntimeError(self._messages)

        super().__init__(data)

    def __getitem__(self, index) -> SExprBase:
        if self._data is None:
            return SExprEmpty()

        if isinstance(index, tuple):
            items = self.all(keys=index)
            if 0 == len(items):
                return SExprEmpty()

            if 1 != len(items):
                raise RuntimeError(
                    f"indexer can only be used on unique keys ('{index}' appears {len(items)} times)")

            return items[0]

        if isinstance(index, str):
            items = [x for x in self._lists if x[0] == index]
            if 0 == len(items):
                return SExprEmpty()

            if 1 != len(items):
                raise RuntimeError(
                    f"indexer can only be used on unique keys ('{index}' appears {len(items)} times)")

            return SExpr(items[0])

        if isinstance(index, int):
            if index >= len(self._atoms):
                return SExprEmpty()

            return self._atoms[index]

        raise RuntimeError(f"bad index type {type(index)}")

    def __setitem__(self, index, value) -> None:
        if self._data is None:
            return

        if isinstance(index, int):
            if isinstance(self._data[index], str):
                self._data[index] = value
                return

            if isinstance(self._data[index], float):
                self._data[index] = value
                return

            if isinstance(self._data[index], int):
                self._data[index] = value
                return

        raise RuntimeError(f"[{index}] has invalid atom type {type(index)}")

    def _allowed_list_tags(self) -> (str):
        allowed_tags = self._required_lists if self._required_lists is not None else ()
        allowed_tags = allowed_tags + \
            (self._optional_lists if self._optional_lists is not None else ())
        return allowed_tags

    def _keyed_list_tags(self) -> (str):
        keyed_tags = [x[:-2]
                      for x in self._allowed_list_tags() if x.endswith('[]')]
        return keyed_tags

    def _is_allowed_list_tag(self, tag: str) -> bool:
        allowed_tags = self._allowed_list_tags()

        if not tag in allowed_tags and not tag + '[]' in allowed_tags:
            return False

        return True

    def _validate(self) -> bool:
        self._messages = []

        result = self._validate_inner()

        # todo foss: logging
        # if not result:
        #     print(self._messages)

        return result

    def _validate_inner(self):

        if 0 == len(self._data):
            return self._error("data has zero length")

        if 0 == len(self._atoms):
            return self._error("atoms has zero lenhth")

        if not isinstance(self.tag, str):
            return self._error('tag is not a string')

        max_number_of_atoms = 0

        if self._required_atoms is not None:
            max_number_of_atoms += len(self._required_atoms)

            if len(self._atoms) < len(self._required_atoms):
                return self._error(f'not enough atoms')

            for i, required_type in enumerate(self._required_atoms):
                if 'number' == required_type:
                    if not isinstance(self._atoms[i], float) and not isinstance(self._atoms[i], int):
                        return self._error(f"atom {i} should be number")

                if 'string' == required_type:
                    if not isinstance(self._atoms[i], str):
                        return self._error(f"atom {i} should be string")

        if self._optional_atoms is not None:
            max_number_of_atoms += len(self._optional_atoms)
            first_index = 0 if self._required_atoms is None else len(
                self._required_atoms)

            if len(self._atoms) > first_index + len(self._optional_atoms):
                return self._error('too many atoms')

            for i, optional_type in enumerate(self._optional_atoms):
                index = first_index + i

                if index < len(self._atoms):
                    atom = self._atoms[index]

                    if 'number' == optional_type:
                        if not isinstance(atom, float) and not isinstance(atom, int):
                            return False

                    if 'string' == optional_type:
                        if not isinstance(atom, str):
                            self._error(
                                f"atom index {i} should be string (found '{atom}')")
                            return False

        if len(self._atoms) > max_number_of_atoms:
            self._error("too many atoms")
            return False

        list_tags = [x[0] for x in self._lists]

        if self._required_lists is not None:
            # todo foss: this test is no good given keyed lists
            if len(self._lists) < len(self._required_lists):
                self._error("not enough required lists")
                return False

        for tag in list_tags:
            if not self._is_allowed_list_tag(tag):
                self._error(f"invalid list tag '{tag}'")
                return False

            tag_count = list_tags.count(tag)
            if tag_count > 1 and not tag in self._keyed_list_tags():
                self._error(
                    f"too many instances of list tag '{tag}' ({tag_count})")
                return False

        if self._required_tag is not None:
            if self.tag != self._required_tag:
                self._error(
                    f"required tag '{self._required_tag}' missing (found '{self.tag}')")
                return False

        return True

    def _validate_or_raise(self, message=None):
        if not self._validate():
            raise RuntimeError(
                "validation failed" if message is None else message)

    def _error(self, message: str) -> bool:
        tag = self.tag if self.tag is not None else '[None]'
        message = f"ERROR: {tag}:{type(self)}:{message}"
        # print(message)  # todo foss: proper logging
        # print(self._data)f
        self._messages.append(message)
        return False

    def _get_keyed_list(self, tag: str, key: str) -> []:
        lists = self._find_lists((tag, key,))

        return lists[0]

    def all(self, keys: Any) -> (SExprBase):
        result = []

        if isinstance(keys, tuple):

            for l in self._lists:
                if len(l) >= len(keys):
                    matches = [keys[a] == l[a] for a in range(len(keys))]

                    if all(matches):
                        result.append(SExpr(l))

        elif isinstance(keys, str):

            for record in self._lists:
                if keys == record[0]:
                    result.append(SExpr(record))

        else:
            raise RuntimeError(f"bad keys type'{type(keys)}'")

        if result is None:
            result.append(SExprEmpty())

        return tuple(result)

    def first(self, keys: Any) -> SExprBase:
        a = self.all(keys)
        return a[0] if len(a) != 0 else SExprEmpty()

    def remove(self, index: Any) -> int:
        if isinstance(index, int):
            if index < len(self._atoms):
                self._remove_atom(self._atoms[index])
        elif isinstance(index, list) or isinstance(index, str):
            for record in self.all(index):
                self._data.remove(record._data)

    def add_list(self, ll: []):
        self._add_list(ll)



import re
import typing

# foss: the clever regex magic below is lifted verbatim from the KiCad repo here:
# https://gitlab.com/kicad/libraries/kicad-library-utils/-/blob/master/common/sexpr.py


term_regex = r"""(?mx)
    \s*(?:
        (?P<brackl>\()|
        (?P<brackr>\))|
        (?P<num>[+-]?\d+\.\d+(?=[\ \)])|\-?\d+(?=[\ \)]))|
        (?P<sq>"([^"]|(?<=\\)")*")|
        (?P<s>[^(^)\s]+)
       )"""

# foss: here I have modified it a bit to make it more readable

# open_bracket_pattern = r"""\("""
# close_bracket_pattern = r"""\)"""
# number_pattern = r"""[+-]?\d+\.\d+(?=[\ \)])|\-?\d+(?=[\ \)])"""
# quoted_string_pattern = r""""([^"]|(?<=\\)")*""""
# string_pattern = r"""[^(^)\s]+"""

# term_regex = fr"""(?mx)
#     \s*(?:
#         (?P<open_bracket>{open_bracket_pattern})|
#         (?P<close_bracket>\))|
#         (?P<number>[+-]?\d+\.\d+(?=[\ \)])|\-?\d+(?=[\ \)]))|
#         (?P<quoted_string>"([^"]|(?<=\\)")*")|
#         (?P<string>[^(^)\s]+)
#        )"""

# foss: ...and the rest is heavily based on same:


def _is_integer(n):
    try:
        int(n)
    except:
        return False

    return True


def _is_numeric(n):
    try:
        int(n)
    except:
        try:
            float(n)
        except:
            return False

    return True


def parse_sexpr(sexp: str) -> typing.Any:
    stack = []
    out = []

    all = re.findall(term_regex, sexp)

    for i, item in enumerate(all):

        if '(' == item[0]:
            stack.append(out)
            out = []
        elif ')' == item[1]:
            assert stack, "Trouble with nesting of brackets"
            tmpout, out = out, stack.pop(-1)
            out.append(tmpout)
        elif '' != item[2]:
            if _is_integer(item[2]):
                out.append(int(item[2]))
            else:
                out.append(float(item[2]))
        elif '' != item[3]:
            out.append(item[3][1:-1].replace(r"\"", '"'))
        elif '' != item[4]:
            out.append(item[4])
        elif '' != item[5]:
            out.append(item[5])
        else:
            pass

    return out[0]


def build_sexpr(exp, level: int = 0) -> str:

    if isinstance(exp, list):
        indent = '  ' * level
        content = " ".join(build_sexpr(x, level=level+1) for x in exp)
        return f"\n{indent}({content})"

    if isinstance(exp, float) or isinstance(exp, int):
        return str(exp)

    if isinstance(exp, str):
        exp = exp.replace("\"", "\\\"")

        if len(exp) == 0 or (' ' in exp) or ('.' in exp) or ('(' in exp) or ('\n' in exp) or _is_numeric(exp):
            exp = f'"{exp}"'

        exp = exp.replace("\n", "\\n")

        return exp

    raise RuntimeError("bad sexpr type")



from .kicad import parse_sexpr, build_sexpr
import typing
from typing import Any


class SExprBase():
    _allow_instantiate_base = False

    allow_empty = True

    def __new__(cls, *args, **kwargs):
        if cls is SExprBase and not SExprBase._allow_instantiate_base:
            raise RuntimeError("please do not instantiate base class")

        return object.__new__(cls)

    def __init__(self, data: []):
        self._messages = []  # todo foss: implement this

        if data is None:
            if not SExprBase.allow_empty:
                raise RuntimeError("created empty instance")

            self._data = None
        elif isinstance(data, list):
            self._data = data
        elif isinstance(data, SExprBase):
            self._data = data._data
        else:
            raise RuntimeError(f"invalid data type '{type(data)}'")

    def __repr__(self) -> str:
        return f"<sexpr>{build_sexpr(self._data)}</sexpr>"

    @property
    def is_empty(self) -> bool:
        return self._data is None

    @property
    def tag(self) -> str:
        if self.is_empty:
            return "<EMPTY>"  # todo foss: what to do in these cases?

        if len(self._atoms) == 0:
            return "<NOTAG>"

        name = self._atoms[0]

        if not isinstance(name, str):
            return None

        return name

    @property
    def _atoms(self) -> (Any):
        if self.is_empty:
            return ()

        l = [x for x in self._data if isinstance(
            x, str) or isinstance(x, float) or isinstance(x, int)]
        return tuple(l)

    @property
    def _lists(self) -> ([Any]):
        if self.is_empty:
            return ()

        l = [x for x in self._data if isinstance(x, list)]
        return tuple(l)

    @property
    def _atom_indicies(self) -> (int):
        ii = ()

        for i, e in enumerate(self._data):
            if not isinstance(e, list):
                ii = ii + (i,)

        return ii

    @staticmethod
    def _is_atom_type(atom):
        if isinstance(atom, str):
            return True

        if isinstance(atom, int):
            return True

        if isinstance(atom, float):
            return True

    def _add_atom(self, atom: Any):

        if not SExprBase._is_atom_type(atom):
            raise RuntimeError(f"invalid atom type '{type(atom)}'")

        self._data.append(atom)

    def _set_atom(self, index, atom):
        if not isinstance(index, int):
            raise RuntimeError("non-integer index")

        if index >= len(self._atoms):
            raise RuntimeError("atom index out of range")

        if not SExprBase._is_atom_type(atom):
            raise RuntimeError(f"invalid atom type '{type(atom)}'")

        index = self._atom_indicies[index]

        self._data[index] = atom

    def _remove_atom(self, atom: Any):
        if atom in self._data:
            self._data.remove(atom)

    def _add_list(self, ll: []):
        if not isinstance(ll, list):
            raise RuntimeError(f"invalid list type '{type(ll)}'")

        if len(ll) < 2:
            raise RuntimeError(f"invalid list length '{len(ll)}'")

        if ll[0] is None:
            raise RuntimeError(f"list name is None")

        if not isinstance(ll[0], str):
            raise RuntimeError(
                f"list name is must be string (is {type(ll[0])}")

        if len(ll[0]) == 0:
            raise RuntimeError(f"list name is zero-length string")

        self._data.append(ll)

    def _remove_list(self, ll: []):
        if not ll in self._lists:
            raise RuntimeError(f"list {ll} is not in lists")

        self._data.remove(ll)

    def _find_lists(self, keys: Any) -> ():

        if isinstance(keys, tuple):
            result = []

            for l in self._lists:
                if len(l) >= len(keys):
                    matches = [keys[a] == l[a] for a in range(len(keys))]

                    if all(matches):
                        result.append(l)
            return tuple(result)

        elif isinstance(keys, str):
            return tuple([x for x in self._lists if keys == x[0]])

        raise RuntimeError(f"invalid key(s) type '{type(keys)}'")



from john import TestCase

import copy

from .sexpr import SExpr, SExprEmpty


class TestInit(TestCase):
    def test_default_init_raises(self):
        with self.assertRaisesAny():
            SExpr()

    def test_init_with_None_raises(self):
        with self.assertRaisesAny():
            SExpr(None)

    def test_init_with_scalar_raises(self):
        with self.assertRaisesAny():
            SExpr('ZZ9PZA')

        with self.assertRaisesAny():
            SExpr(('ZZ9PZA',))

        with self.assertRaisesAny():
            SExpr(42)

        with self.assertRaisesAny():
            SExpr(3.14)

    def test_init_with_list_does_not_raise(self):
        with self.assertDoesNotRaise():
            SExpr([])

    def test_init_with_SExpr_does_not_raise(self):
        SExpr(SExpr([]))
        with self.assertDoesNotRaise():
            SExpr(SExpr([]))

    def test_init_with_required_atoms_does_not_raise(self):
        with self.assertDoesNotRaise():
            SExpr([], required_atoms=())

    def test_required_atoms_must_be_tuple(self):
        with self.assertRaisesAny():
            SExpr([], required_atoms=[])

    def test_init_with_optional_atoms_does_not_raise(self):
        with self.assertDoesNotRaise():
            SExpr([], optional_atoms=())

    def test_optional_atoms_must_be_tuple(self):
        with self.assertRaisesAny():
            SExpr([], optional_atoms=[])

    def test_init_with_required_lists_does_not_raise(self):
        with self.assertDoesNotRaise():
            SExpr([], required_lists=())

    def test_required_lists_must_be_tuple(self):
        with self.assertRaisesAny():
            SExpr([], required_lists=[])

    def test_required_lists_must_contain_strings(self):
        with self.assertRaisesAny():
            SExpr([], required_lists=('A', 'B', 3.14))

    def test_init_with_optional_lists_does_not_raise(self):
        with self.assertDoesNotRaise():
            SExpr([], optional_lists=())

    def test_optional_lists_must_be_tuple(self):
        with self.assertRaisesAny():
            SExpr([], optional_lists=3.14)

    def test_optional_lists_must_contain_strings(self):
        with self.assertRaisesAny():
            SExpr([], optional_lists=('A', [], 'B'))


class TestValidate(TestCase):
    def test_data_must_not_be_empty(self):
        sut = SExpr([])

        self.assertFalse(sut._validate())

    def test_must_not_have_zero_atoms(self):
        sut = SExpr([['ZZ9PZA']])

        self.assertFalse(sut._validate())

    def test_first_atom_must_be_string(self):
        self.assertFalse(SExpr([42])._validate())
        self.assertFalse(SExpr([3.14])._validate())

    def test_invalid_atom_type_raises(self):
        with self.assertRaisesAny():
            SExpr(['ZZ9PZA'], required_atoms=('a',))

    def test_string_atom_type_does_not_raise(self):
        with self.assertDoesNotRaise():
            SExpr(['ZZ9PZA'], required_atoms=('string',))

    def test_number_atom_type_does_not_raise(self):
        with self.assertDoesNotRaise():
            SExpr(['ZZ9PZA'], required_atoms=('number',))

    def test_first_atom_string_wrong_type(self):
        self.assertFalse(
            SExpr(['ZZ9PZA'], required_atoms=('number',))._validate())

    def test_first_atom_string(self):
        self.assertTrue(
            SExpr(['ZZ9PZA'], required_atoms=('string',))._validate())

    def test_second_atom_string(self):
        self.assertTrue(SExpr(['ZZ9PZA', 'A'], required_atoms=(
            'string', 'string',))._validate())

    def test_second_atom_number_is_float(self):
        self.assertTrue(SExpr(['ZZ9PZA', 3.14], required_atoms=(
            'string', 'number',))._validate())

    def test_second_atom_number_is_int(self):
        self.assertTrue(SExpr(['ZZ9PZA', 1729], required_atoms=(
            'string', 'number',))._validate())

    def test_more_atoms_than_required(self):
        self.assertFalse(SExpr(['ZZ9PZA', 'A', 'B'], required_atoms=(
            'string', 'string',))._validate())

    def test_too_many_optional_atoms(self):
        self.assertFalse(SExpr(['ZZ9PZA', 'A', 'B'], required_atoms=(
            'string', 'string',), optional_atoms=())._validate())

    def test_one_optional_number_atom(self):
        self.assertTrue(SExpr(['ZZ9PZA', 'A', 3.14], required_atoms=(
            'string', 'string',), optional_atoms=('number',))._validate())

    def test_one_optional_number_atom_wrong_type(self):
        self.assertFalse(SExpr(['ZZ9PZA', 'A', 'B'], required_atoms=(
            'string', 'string',), optional_atoms=('number',))._validate())

    def test_one_optional_string_atom_wrong_type(self):
        self.assertFalse(SExpr(['ZZ9PZA', 'A', 3.14], required_atoms=(
            'string', 'string',), optional_atoms=('string',))._validate())

    def test_one_optional_atom_not_supplied(self):
        SExpr(['ZZ9PZA', 'A'], required_atoms=('string', 'string',),
              optional_atoms=('number',))._validate()
        self.assertTrue(SExpr(['ZZ9PZA', 'A'], required_atoms=(
            'string', 'string',), optional_atoms=('number',))._validate())

    def test_two_optional_atoms(self):
        self.assertTrue(SExpr(['ZZ9PZA', 'A', 'B', 'C'], required_atoms=(
            'string', 'string',), optional_atoms=('string', 'string',))._validate())

    def test_required_lists(self):
        self.assertTrue(SExpr(['ZZ9PZA', ['A']], required_atoms=(
            'string',), required_lists=('A',))._validate())

    def test_required_lists_longer_than_lists(self):
        self.assertFalse(SExpr(['ZZ9PZA', ['A']], required_atoms=(
            'string',), required_lists=('A', 'B',))._validate())

    def test_required_lists_missing(self):
        self.assertFalse(SExpr(['ZZ9PZA', ['B']], required_atoms=(
            'string',), required_lists=('A',))._validate())

    def test_required_lists_missing_one(self):
        self.assertFalse(SExpr(['ZZ9PZA', ['A'], ['B']], required_atoms=(
            'string',), required_lists=('A', 'C',))._validate())

    def test_one_optional_list_missing(self):
        self.assertFalse(SExpr(['ZZ9PZA', ['A'], ['B']], required_atoms=(
            'string',), required_lists=('A',), optional_lists=('C',))._validate())

    def test_one_optional_list(self):
        self.assertTrue(SExpr(['ZZ9PZA', ['A'], ['B']], required_atoms=(
            'string',), required_lists=('A',), optional_lists=('B',))._validate())

    def test_two_optional_lists_second_missing(self):
        self.assertFalse(SExpr(['ZZ9PZA', ['A'], ['B'], ['C']], required_atoms=(
            'string',), required_lists=('A',), optional_lists=('B',))._validate())

    def test_required_and_optional_lists_mixed(self):
        self.assertTrue(SExpr(['ZZ9PZA', ['opt_A'], ['req_B'], ['opt_C']], required_atoms=(
            'string',), required_lists=('req_B',), optional_lists=('opt_A', 'opt_C'))._validate())

    def test_only_optional_lists(self):
        self.assertTrue(SExpr(['ZZ9PZA', ['opt_A'], ['opt_B']], required_atoms=(
            'string',), optional_lists=('opt_A', 'opt_B'))._validate())

    def test_keyed_list_accepted(self):
        self.assertTrue(SExpr(['ZZ9PZA', ['A', 'key_A']], required_atoms=(
            'string',),  required_lists=('A[]',))._validate())

    def test_non_keyed_list_does_not_allow_duplicates(self):
        self.assertFalse(SExpr(['ZZ9PZA', ['A'], ['A']], required_atoms=(
            'string',),  required_lists=('A',))._validate())

    def test_keyed_list_does_allow_duplicates(self):
        self.assertTrue(SExpr(['ZZ9PZA', ['A'], ['A']], required_atoms=(
            'string',),  required_lists=('A[]',))._validate())

    def test_required_tag_wrong_tag(self):
        self.assertFalse(SExpr(['ZZ9PZA'], required_atoms=(
            'string',), required_tag='A')._validate())

    def test_required_tag(self):
        self.assertTrue(SExpr(['ZZ9PZA'], required_atoms=(
            'string',), required_tag='ZZ9PZA')._validate())

    def test_validate_or_raise(self):
        with self.assertRaisesAny():
            SExpr(['ZZ9PZA'], required_atoms=('string',),
                  required_tag='A')._validate_or_raise()


class TestAtomIndexer(TestCase):
    def test_atom_indexer_not_found_does_not_raise(self):
        sut = SExpr(['ZZ9PZA', 'other'])

        with self.assertDoesNotRaise():
            sut[2]

    def test_atom_indexer_not_found_returns_Empty(self):
        sut = SExpr(['ZZ9PZA', 'other'])

        actual = sut[2]

        self.assertIsInstance(actual, SExprEmpty)

    def test_atom_get_indexer(self):
        expected = 'ZZ9PZA'

        sut = SExpr(['blah', expected, 'other'])

        actual = sut[1]

        self.assertEqual(actual, expected)

    def test_set_indexer_non_integer_index_raises(self):
        sut = SExpr(['ZZ9PZA'])

        with self.assertRaisesAny():
            sut['Arthur'] = 1

    def test_atom_set_indexer_int(self):
        sut = SExpr(['ZZ9PZA'])

        expected = 42

        sut[0] = expected

        actual = sut[0]

        self.assertEqual(actual, expected)

    def test_atom_set_indexer_float(self):
        sut = SExpr(['ZZ9PZA'])

        expected = 3.14

        sut[0] = expected

        actual = sut[0]

        self.assertEqual(actual, expected)

    def test_atom_set_indexer_string(self):
        sut = SExpr(['ZZ9PZA'])

        expected = 'Ford'

        sut[0] = expected

        actual = sut[0]

        self.assertEqual(actual, expected)


class TestListIndexer(TestCase):
    def test_list_indexer_not_found_does_not_raise(self):
        sut = SExpr(['ZZ9PZA', 'other'])

        with self.assertDoesNotRaise():
            sut['cabbage']

    def test_list_indexer_not_found_returns_Empty(self):
        sut = SExpr(['ZZ9PZA', 'other'])

        actual = sut['cabbage']

        self.assertIsInstance(actual, SExprEmpty)

    def test_list_get_indexer(self):
        key = 'ZZ9PZA'

        expected = [key, 'blah']

        sut = SExpr(['blah', expected, ['other']])

        actual = sut[key]
        self.assertIsInstance(actual, SExpr)
        self.assertListEqual(actual._data, expected)

    def test_list_get_indexer_list_key_raises(self):
        sut = SExpr(['ZZ9PZA'])

        with self.assertRaisesAny():
            sut[[]]

    def test_list_get_indexer_tuple_key_not_found_returns_Empty(self):
        keys = ('ZZ9PZA', 'Lemon')

        expected = [keys[0], keys[1], 'blah']

        sut = SExpr(['blah', expected, ['other']])

        actual = sut[tuple(keys) + ('a',)]

        self.assertIsInstance(actual, SExprEmpty)

    def test_list_get_indexer_tuple_key(self):
        keys = ('ZZ9PZA', 'Lemon')

        expected = [keys[0], keys[1], 'blah']

        sut = SExpr(['blah', expected, ['other']])

        actual = sut[tuple(keys)]

        self.assertIsInstance(actual, SExpr)
        self.assertListEqual(actual._data, expected)


class TestAddList(TestCase):
    def test_add_list(self):
        sut = SExpr(['ZZ9PZA'])

        ll = ['a', 'b']

        sut.add_list(ll)

        expected = copy.deepcopy(ll)

        actual = sut._lists[0]

        self.assertListEqual(actual, expected)


class TestAll(TestCase):
    def test_empty_all_returns_empty_tuple(self):
        sut = SExpr(['A'])

        actual = sut.all('ZZ9PZA')

        expected = ()

        self.assertTupleEqual(actual, expected)

    def test_all_returns_tuple(self):
        sut = SExpr(['ZZ9PZA'])

        actual = sut.all('ZZ9PZA')

        self.assertIsInstance(actual, tuple)

    def test_all(self):
        data = ['ZZ9PZA',
                ['A', 'B', 'C'],
                ['B', 'B', 'D'],
                ['A', 'B', 'E']
                ]

        sut = SExpr(data)

        expected = [data[1], data[3]]

        actual = [sexpr._data for sexpr in sut.all(('A', 'B',))]

        self.assertSequenceEqual(expected, actual)


class TestFirst(TestCase):
    def test_missing_first_returns_Empty(self):
        sut = SExpr(['A'])

        actual = sut.first('ZZ9PZA')

        self.assertIsInstance(actual, SExprEmpty)

    def test_first(self):
        sut = SExpr(['ZZ9PZA', ['ZZ9PZA', 'A'], ['ZZ9PZA', 'B']])

        actual = sut.first('ZZ9PZA')._data

        expected = sut._lists[0]

        self.assertListEqual(actual, expected)


class TestRemove(TestCase):
    def test_empty_remove_atom_does_not_raise(self):
        with self.assertDoesNotRaise():
            SExpr(['ZZ9PZA']).remove(2)

    def test_remove_atom(self):
        sut = SExpr(['A', 'B', 'C'])

        expected = (sut[0], sut[2])

        sut.remove(1)

        actual = sut._atoms

        self.assertTupleEqual(actual, expected)

    def test_remove_list(self):
        sut = SExpr(['ZZ9PZA', ['A', 'B'], ['C', 'D']])

        expected = (sut._lists[0],)

        sut.remove('C')

        actual = sut._lists

        self.assertTupleEqual(actual, expected)


class TestKeyed_lists(TestCase):
    def test_get_keyed_list(self):
        expected = ['A', '2']
        sut = SExpr(['ZZ9PZA', ['A', '1'], expected],
                    required_atoms=('string',), required_lists=('A[]',))

        actual = sut._get_keyed_list('A', '2')

        self.assertSequenceEqual(actual, expected)
